"""
NER for LinCE dataset
"""
import numpy as np
from .ner import tokenize_and_align_labels, preprocess_token
from datasets import load_dataset, load_metric
from seqeval.metrics import f1_score
from transformers import DataCollatorForTokenClassification, AutoModelForTokenClassification
from ..training import train_model

metric = load_metric("seqeval")

"""
TODO: this is a hack to make it work with the current version of seqeval
As you can see, I add "B-" to each label because it only works with BIO tags
"""

id2label =[
    'B-VERB',
    'B-PUNCT',
    'B-PRON',
    'B-NOUN',
    'B-DET',
    'B-ADV',
    'B-ADP',
    'B-INTJ',
    'B-CONJ',
    'B-ADJ',
    'B-AUX',
    'B-SCONJ',
    'B-PART',
    'B-PROPN',
    'B-NUM',
    'B-UNK',
    'B-X',
]

label2id = {v:k for k,v in enumerate(id2label)}


def load_datasets(lang="es", preprocess=True):
    """
    Load NER datasets
    """

    lince = load_dataset("lince", "pos_spaeng")

    # This hack is because of seqeval only working with BIO tags

    lince = lince.map(
        lambda x: {"pos": ["B-"+x for x in x["pos"]]}
    )
    """
    TODO: None is for test labels which are not available
    """
    lince = lince.map(
        lambda x: {"labels": [label2id.get(x, None) for x in x["pos"]]}
    )

    if preprocess:
        lince = lince.map(
            lambda x: {
                "words": [preprocess_token(word, lang) for word in x["words"]]
            }
        )

    return lince["train"], lince["validation"], lince["test"]

metric = load_metric("seqeval")

def compute_metrics(eval_preds):
    """
    Compute metrics for POS
    """
    logits, labels = eval_preds
    predictions = np.argmax(logits, axis=-1)

    # Remove ignored index (special tokens) and convert to labels
    true_labels = [[id2label[l] for l in label if l != -100] for label in labels]
    true_predictions = [
        [id2label[p] for (p, l) in zip(prediction, label) if l != -100]
        for prediction, label in zip(predictions, labels)
    ]
    all_metrics = metric.compute(predictions=true_predictions, references=true_labels)
    ret = {
        "precision": all_metrics["overall_precision"],
        "recall": all_metrics["overall_recall"],
        "macro_f1": f1_score(true_labels, true_predictions, average="macro"),
        "micro_f1": f1_score(true_labels, true_predictions, average="micro"),
        "accuracy": all_metrics["overall_accuracy"],
    }

    for k, v in all_metrics.items():
        if not k.startswith("overall"):
            ret[k + "_f1"] = v["f1"]
            ret[k + "_precision"] = v["precision"]
            ret[k + "_recall"] = v["recall"]

    return ret

def train(
    base_model, lang, epochs=5,
    metric_for_best_model="accuracy",
    **kwargs):

    train_dataset, dev_dataset, test_dataset = load_datasets(
        lang=lang
    )

    return train_model(
        base_model,
        train_dataset=train_dataset, dev_dataset=dev_dataset, test_dataset=dev_dataset,
        id2label=id2label, lang=lang, epochs=epochs,
        # Custom stuff for this thing to work
        tokenize_fun=tokenize_and_align_labels,
        auto_class=AutoModelForTokenClassification,
        data_collator_class=DataCollatorForTokenClassification,
        metrics_fun=compute_metrics,
        metric_for_best_model=metric_for_best_model,
        **kwargs
    )


import torch
from .preprocessing import preprocess_tweet
from transformers import (
    AutoTokenizer, AutoModelForSequenceClassification, AutoModelForTokenClassification,
    DataCollatorWithPadding,
    Trainer, TrainingArguments,
    pipeline
)
from datasets import Dataset
from torch.nn import functional as F


models = {
    "es": {
        "sentiment": {
            "model_name": "pysentimiento/robertuito-sentiment-analysis",
        },
        "emotion": {
            "model_name": "pysentimiento/robertuito-emotion-analysis",
        },
        "hate_speech": {
            "model_name": "pysentimiento/robertuito-hate-speech",
        },
        "ner": {
            "model_name": "pysentimiento/robertuito-ner",
        },
        "pos": {
            "model_name": "pysentimiento/robertuito-pos",
        }
    },
    "en": {
        "sentiment": {
            "model_name": "finiteautomata/bertweet-base-sentiment-analysis",
            # BerTweet uses different preprocessing args
            "preprocessing_args": {"user_token": "@USER", "url_token": "HTTPURL"}
        },
        "emotion": {
            "model_name": "finiteautomata/bertweet-base-emotion-analysis",
            "preprocessing_args": {"user_token": "@USER", "url_token": "HTTPURL"}
        },
        "hate_speech": {
            "model_name": "pysentimiento/bertweet-hate-speech",
            "preprocessing_args": {"user_token": "@USER", "url_token": "HTTPURL"}
        },
        "ner": {
            "model_name": "pysentimiento/robertuito-ner",
        }
    },
}


class AnalyzerOutput:
    """
    Base class for classification output
    """

    def __init__(self, sentence, probas, is_multilabel=False):
        """
        Constructor
        """
        self.sentence = sentence
        self.probas = probas
        self.is_multilabel = is_multilabel
        if not is_multilabel:
            self.output = max(probas.items(), key=lambda x: x[1])[0]
        else:
            self.output = [
                k for k, v in probas.items() if v > 0.5
            ]

    def __repr__(self):
        ret = f"{self.__class__.__name__}"
        if not self.is_multilabel:
            formatted_probas = sorted(self.probas.items(), key=lambda x: -x[1])
        else:
            formatted_probas = list(self.probas.items())
        formatted_probas = [f"{k}: {v:.3f}" for k, v in formatted_probas]
        formatted_probas = "{" + ", ".join(formatted_probas) + "}"
        ret += f"(output={self.output}, probas={formatted_probas})"

        return ret


class BaseAnalyzer:
    def __init__(self, model, tokenizer, task, preprocessing_args={}, batch_size=32):
        """
        Constructor for SentimentAnalyzer class

        Arguments:

        model_name: str or path
            Model name or
        """
        self.model = model
        self.tokenizer = tokenizer
        self.preprocessing_args = preprocessing_args
        self.batch_size = batch_size
        self.task = task

        self.tokenizer.model_max_length = 128
        self.problem_type = self.model.config.problem_type
        self.id2label = self.model.config.id2label

        self.eval_trainer = Trainer(
            model=self.model,
            args=TrainingArguments(
                output_dir=".",
                per_device_eval_batch_size=batch_size,
            ),
            data_collator=DataCollatorWithPadding(
                self.tokenizer, padding="longest"),
        )

    def _tokenize(self, batch):
        return self.tokenizer(
            batch["text"], padding=False, truncation=True
        )


class AnalyzerForSequenceClassification(BaseAnalyzer):
    """
    Wrapper to use sentiment analysis models as black-box
    """

    @classmethod
    def from_model_name(cls, model_name, task, preprocessing_args={}, batch_size=32, **kwargs):
        """
        Constructor for SentimentAnalyzer class

        Arguments:

        model_name: str or path
            Model name or
        """
        model = AutoModelForSequenceClassification.from_pretrained(model_name)
        tokenizer = AutoTokenizer.from_pretrained(model_name)
        return cls(model, tokenizer, task, preprocessing_args, batch_size, **kwargs)

    def _get_output(self, sentence, logits):
        """
        Get output from logits

        It takes care of the type of problem: single or multi label classification
        """
        if self.problem_type == "multi_label_classification":
            is_multilabel = True
            probs = torch.sigmoid(logits).view(-1)
        else:
            is_multilabel = False
            probs = torch.softmax(logits, dim=1).view(-1)

        probas = {self.id2label[i]: probs[i].item() for i in self.id2label}
        return AnalyzerOutput(sentence, probas=probas, is_multilabel=is_multilabel)

    def _predict_single(self, sentence):
        """
        Predict single

        Do it this way (without creating dataset) to make it faster
        """
        device = self.eval_trainer.args.device
        sentence = preprocess_tweet(sentence, **self.preprocessing_args)
        idx = torch.LongTensor(
            self.tokenizer.encode(
                sentence, truncation=True,
                max_length=self.tokenizer.model_max_length,
            )
        ).view(1, -1).to(device)
        output = self.model(idx)
        logits = output.logits
        return self._get_output(sentence, logits)

    def predict(self, inputs):
        """
        Return most likely class for the sentence

        Arguments:
        ----------
        inputs: string or list of strings
            A single or a list of sentences to be predicted

        Returns:
        --------
            List or single output with probabilities
        """

        # If single string => predict it single
        if isinstance(inputs, str):
            return self._predict_single(inputs)

        sentences = [
            preprocess_tweet(sent, **self.preprocessing_args) for sent in inputs
        ]
        dataset = Dataset.from_dict({"text": sentences})
        dataset = dataset.map(self._tokenize, batched=True,
                              batch_size=self.batch_size)

        output = self.eval_trainer.predict(dataset)
        logits = torch.tensor(output.predictions)
        rets = [self._get_output(sent, logits_row.view(1, -1))
                for sent, logits_row in zip(sentences, logits)]

        return rets


class AnalyzerForTokenClassification(BaseAnalyzer):
    @classmethod
    def from_model_name(cls, model_name, task, preprocessing_args={}, batch_size=32, **kwargs):
        """
        Constructor for SentimentAnalyzer class

        Arguments:

        model_name: str or path
            Model name or
        """
        model = AutoModelForTokenClassification.from_pretrained(model_name)
        tokenizer = AutoTokenizer.from_pretrained(model_name)
        return cls(model, tokenizer, task, preprocessing_args, batch_size, **kwargs)

    def __init__(self, model, tokenizer, task, preprocessing_args={}, batch_size=32, aggregation_strategy="simple"):
        super().__init__(model, tokenizer, task, preprocessing_args, batch_size)
        self.pipeline = pipeline("ner", model=model, tokenizer=tokenizer)
        self.aggregation_strategy = aggregation_strategy

    def predict(self, inputs):
        """
        Predict token classification
        """
        if isinstance(inputs, str):
            inputs = [inputs]

        sentences = [
            preprocess_tweet(sent, **self.preprocessing_args) for sent in inputs
        ]
        ret = self.pipeline(
            sentences, aggregation_strategy=self.aggregation_strategy)

        for sent, entities in zip(sentences, ret):
            for entity in entities:
                start, end = entity["start"], entity["end"]
                entity["text"] = sent[start:end].strip()
                entity["type"] = entity.pop("entity_group")

        if len(sentences) == 1:
            return ret[0]
        return ret


def create_analyzer(task=None, lang=None, model_name=None, preprocessing_args={}, **kwargs):
    """
    Create analyzer for the given task

    Arguments:
    ----------
    task: str
        Task name (sentiment or emotion)
    lang: str
        Language code (es or en)
    model_name: str
        Model name or path
    preprocessing_args: dict
        Preprocessing arguments

    Returns:
    --------
        SentimentAnalyzer or EmotionAnalyzer
    """
    if not (model_name or (lang and task)):
        raise ValueError("model_name or (lang and task) must be provided")

    preprocessing_args = preprocessing_args or {}
    if task in {"ner", "pos"}:
        analyzer_class = AnalyzerForTokenClassification
    else:
        analyzer_class = AnalyzerForSequenceClassification

    if not model_name:
        if lang not in models:
            raise ValueError(
                f"Language {lang} not supported -- only supports {models.keys()}")

        if task not in models[lang]:
            raise ValueError(
                f"Task {task} not supported for {lang} -- only supports {models[lang].keys()}")

        model_info = models[lang][task]
        model_name = model_info["model_name"]
        preprocessing_args.update(model_info.get("preprocessing_args", {}))

    preprocessing_args["lang"] = lang
    return analyzer_class.from_model_name(model_name, task, preprocessing_args, **kwargs)



"""
NER for LinCE dataset
"""
from emoji import emoji_lis, demojize
import numpy as np
from seqeval.metrics import f1_score
from datasets import load_dataset, load_metric, Dataset
from transformers import DataCollatorForTokenClassification, AutoModelForTokenClassification
from ..preprocessing import preprocess_tweet
from ..training import train_model

metric = load_metric("seqeval")

id2label = [
    'O',
    'B-EVENT',
    'I-EVENT',
    'B-GROUP',
    'I-GROUP',
    'B-LOC',
    'I-LOC',
    'B-ORG',
    'I-ORG',
    'B-OTHER',
    'I-OTHER',
    'B-PER',
    'I-PER',
    'B-PROD',
    'I-PROD',
    'B-TIME',
    'I-TIME',
    'B-TITLE',
    'I-TITLE',
]

label2id = {v:k for k,v in enumerate(id2label)}


def align_labels_with_tokens(labels, word_ids, ignore_label=-100, label_subwords=False):
    """
    Taken from https://huggingface.co/course/chapter7/2?fw=pt

    Arguments:

    labels: List[int]
        list of NER labels ()

    word_ids: List[int]
        list of word ids (one for each token)

    ignore_label: int (default: -100)
        Value to set to ignored labels ([CLS], [SEP], [PAD] and subwords if label_subwords=False)

    label_subwords: bool (default: True)
        If True, then labels are assigned to the subword that contains the token.

        For instance, if "Bigsubword" (split as Big ##sub ##word )
        is labeled as "B-PER", then "Big" is labeled as "B-PER",
        ##sub ##word as "I-PER" and "##word" is labeled as "I-PER".

        If False, sets to ignore_id the subwords


    """
    new_labels = []
    current_word = None

    if all(l is None for l in labels):
        # All labels are none => test dataset
        return [None] * len(word_ids)

    for word_id in word_ids:
        if word_id != current_word:
            # Start of a new word!
            current_word = word_id
            label = ignore_label if word_id is None else labels[word_id]
            new_labels.append(label)
        elif word_id is None:
            # Special token
            new_labels.append(ignore_label)
        else:
            """
            Same word as previous token
            """
            if label_subwords:
                label = labels[word_id]
                # If the label is B-XXX we change it to I-XXX
                # WARNING: this strategy is only valid for BIO labels
                if label % 2 == 1:
                    label += 1
            else:
                label = ignore_label
            new_labels.append(label)

    return new_labels



def compute_metrics(eval_preds):
    """
    Compute metrics for NER
    """
    logits, labels = eval_preds
    predictions = np.argmax(logits, axis=-1)

    # Remove ignored index (special tokens) and convert to labels
    true_labels = [[id2label[l] for l in label if l != -100] for label in labels]
    true_predictions = [
        [id2label[p] for (p, l) in zip(prediction, label) if l != -100]
        for prediction, label in zip(predictions, labels)
    ]
    all_metrics = metric.compute(predictions=true_predictions, references=true_labels)
    ret = {
        "precision": all_metrics["overall_precision"],
        "recall": all_metrics["overall_recall"],
        "macro_f1": f1_score(true_labels, true_predictions, average="macro"),
        "micro_f1": f1_score(true_labels, true_predictions, average="micro"),
        "accuracy": all_metrics["overall_accuracy"],
    }

    for k, v in all_metrics.items():
        if not k.startswith("overall"):
            ret[k + "_f1"] = v["f1"]
            ret[k + "_precision"] = v["precision"]
            ret[k + "_recall"] = v["recall"]

    return ret

def tokenize_and_align_labels(examples, tokenizer):
    """
    Tokenize examples and also realign labels
    """
    tokenized_inputs = tokenizer(
        examples["words"], truncation=True, is_split_into_words=True
    )
    all_labels = examples["labels"]
    new_labels = []
    word_ids = [tokenized_inputs.word_ids(i) for i in range(len(all_labels))]
    for (labels, wids) in zip(all_labels, word_ids):
        new_labels.append(
            align_labels_with_tokens(labels, wids)
        )

    tokenized_inputs["labels"] = new_labels
    tokenized_inputs["word_ids"] = word_ids

    return tokenized_inputs

def preprocess_token(t, lang, demoji=True, preprocess_hashtags=False, **kwargs):
    """
    Preprocess each token
    """
    token = None
    if t.startswith("http") and "t.co" in t:
        """
        TODO: this is a patch for preprocess_tweet, but it should be fixed in the future
        """
        token = "url"
    else:
        if demoji:
            emojis = emoji_lis(t, language=lang)
            if emojis:
                """
                Put special token
                """
                token = "emoji"

        if not token:
            if len(t) == 1 and not t.isascii():
                token = "."
            else:
                token = preprocess_tweet(
                    t, lang=lang, demoji=False, preprocess_hashtags=preprocess_hashtags,
                    char_replace=False, **kwargs
                )

    if not token:
        """
        token is empty or None => put a placeholder
        """
        token = "."


    return token

def load_conll(path, lang="es"):
    """
    Loads CoNLL-2003 dataset
    """
    with open(path) as f:
        lines = f.read().splitlines()
    data = []
    current_line = []
    for line in lines:
        line = line.strip()
        if line.startswith("# "):
            continue
        elif line == "":
            data.append(current_line)
            current_line = []
        else:
            current_line.append(line.split("\t"))
    return data


def load_dataset_from_conll(path):
    data = load_conll(path)
    words = [[x[0] for x in sentence] for sentence in data]
    langs = [[x[1] for x in sentence] for sentence in data]
    if len(data[0][0]) == 2:
        ner = [[None] * len(x) for x in data]
    else:
        ner = [[x[2] for x in sentence] for sentence in data]

    ## Sanity Check
    for w, l in zip(words, ner):
        assert len(w) == len(l)

    return Dataset.from_dict({
        "words": words,
        "lang": langs,
        "ner": ner,
    })


def load_datasets(lang="es", preprocess=True):
    """
    Load NER datasets
    """

    lince_ner = load_dataset("lince", "ner_spaeng")

    """
    TODO: None is for test labels which are not available
    """

    lince_ner = lince_ner.map(
        lambda x: {"labels": [label2id.get(x, None) for x in x["ner"]]}
    )

    if preprocess:
        lince_ner = lince_ner.map(
            lambda x: {
                "words": [preprocess_token(word, lang) for word in x["words"]]
            }
        )

    return lince_ner["train"], lince_ner["validation"], lince_ner["test"]

def train(
    base_model, lang, epochs=5,
    metric_for_best_model="micro_f1",
    **kwargs):

    train_dataset, dev_dataset, test_dataset = load_datasets(
        lang=lang
    )

    return train_model(
        base_model,
        train_dataset=train_dataset, dev_dataset=dev_dataset, test_dataset=dev_dataset,
        id2label=id2label, lang=lang, epochs=epochs,
        # Custom stuff for this thing to work
        tokenize_fun=tokenize_and_align_labels,
        auto_class=AutoModelForTokenClassification,
        data_collator_class=DataCollatorForTokenClassification,
        metrics_fun=compute_metrics,
        metric_for_best_model=metric_for_best_model,
        **kwargs
    )


from numpy import int64
import pandas as pd
import os
import pathlib
from datasets import Dataset, Value, ClassLabel, Features
from sklearn.model_selection import train_test_split
from .preprocessing import preprocess_tweet

"""
Lo pongo así por huggingface
"""
id2label = {0: 'NEG', 1: 'NEU', 2: 'POS'}
label2id = {v:k for k,v in id2label.items()}

project_dir = pathlib.Path(os.path.dirname(__file__)).parent
data_dir = os.path.join(project_dir, "data")
semeval_dir = os.path.join(data_dir, "SemEval2017")


def load_df(path):

    df = pd.read_csv(path, encoding="latin-1")

    matches = {"NEG": "negative", "NEU": "neutral", "POS": "positive"}

    for label, idx in label2id.items():
        replacement = matches[label]
        df.loc[df["label"] == replacement, "label"] = idx

    df["label"] = df["label"].astype(int)
    return df

def load_datasets(seed=2021, preprocessing_args={}, **kwargs):
    """
    Return train, dev, test datasets
    """
    train_df = load_df(
        os.path.join(semeval_dir, "train.csv")
    )

    test_df = load_df(
        os.path.join(semeval_dir, "test.csv")
    )

    train_df, dev_df = train_test_split(train_df, test_size=0.2)

    print(len(train_df), len(dev_df), len(test_df))

    """
    Tokenize tweets
    """

    en_preprocess = lambda x: preprocess_tweet(x, lang="en", **preprocessing_args)

    train_df["text"] = train_df["text"].apply(en_preprocess)
    dev_df["text"] = dev_df["text"].apply(en_preprocess)
    test_df["text"] = test_df["text"].apply(en_preprocess)

    features = Features({
        'id': Value('int64'),
        'text': Value('string'),
        'label': ClassLabel(num_classes=3, names=["NEG", "NEU", "POS"])
    })


    columns = ["text", "id", "label"]

    train_dataset = Dataset.from_pandas(train_df[columns], features=features)
    dev_dataset = Dataset.from_pandas(dev_df[columns], features=features)
    test_dataset = Dataset.from_pandas(test_df[columns], features=features)

    return train_dataset, dev_dataset, test_dataset


import pandas as pd
import os
import pathlib
from glob import glob
from datasets import Dataset, Value, ClassLabel, Features
from .preprocessing import preprocess_tweet

"""
Lo pongo así por huggingface
"""
id2label = {0: 'NEG', 1: 'NEU', 2: 'POS'}
label2id = {v:k for k,v in id2label.items()}

project_dir = pathlib.Path(os.path.dirname(__file__)).parent
data_dir = os.path.join(project_dir, "data")
tass_dir = os.path.join(data_dir, "sentiment")

def get_lang(file):
    """
    Get language of TASS dataset
    """
    return os.path.splitext(os.path.basename(file))[0]

def load_df(path, test=False):
    """
    Load TASS dataset
    """
    dialect = get_lang(path)

    if test:
        df = pd.read_table(path, names=["id", "text"], index_col=0)
        label_path = os.path.join(
            os.path.dirname(path),
            "labels",
            f"{dialect.lower()}.tsv"
        )

        labels = pd.read_table(label_path, names=["id", "label"], index_col=0)

        df["polarity"] = labels["label"]
    else:
        df = pd.read_table(path, names=["id", "text", "polarity"], index_col=0)

    df["lang"] = dialect

    for label, idx in label2id.items():
        polarity_label = {"NEG": "N", "NEU": "NEU", "POS": "P"}[label]
        df.loc[df["polarity"] == polarity_label, "label"] = idx

    return df


def load_datasets(preprocessing_args={}, preprocess=True, return_df=False, **kwargs):
    """
    Return train, dev, test datasets
    """

    df = pd.read_csv(os.path.join(tass_dir, "tass.csv"))

    df["label"] = df["polarity"].apply(lambda x: label2id[x])

    train_df = df[df["split"] == "train"].copy()
    dev_df = df[df["split"] == "dev"].copy()
    test_df = df[df["split"] == "test"].copy()

    print(len(train_df), len(dev_df), len(test_df))

    """
    Tokenize tweets
    """

    if preprocess:
        preprocess_with_args = lambda x: preprocess_tweet(x, **preprocessing_args)

        train_df["text"] = train_df["text"].apply(preprocess_with_args)
        dev_df["text"] = dev_df["text"].apply(preprocess_with_args)
        test_df["text"] = test_df["text"].apply(preprocess_with_args)

    if return_df:
        return train_df, dev_df, test_df

    features = Features({
        'text': Value('string'),
        'lang': Value('string'),
        'label': ClassLabel(num_classes=3, names=["neg", "neu", "pos"])
    })

    columns = ["text", "lang", "label"]

    train_dataset = Dataset.from_pandas(train_df[columns], features=features)
    dev_dataset = Dataset.from_pandas(dev_df[columns], features=features)
    test_dataset = Dataset.from_pandas(test_df[columns], features=features)

    return train_dataset, dev_dataset, test_dataset



import torch
import pytorch_lightning as pl
import fasttext
import unidecode
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence
from datasets import concatenate_datasets
from torchtext.data.utils import get_tokenizer
from pysentimiento.baselines.models import FFNModel, RNNModel
from pysentimiento.baselines.utils import build_vocab, build_embedding_matrix

class EvalResults(object):
    def __init__(self, predictions, metrics):
        self.predictions = predictions
        self.metrics = metrics

    def __str__(self):
        return f"\n{self.metrics}"


def tokenize_dataset(dataset, tokenizer, vocab):
    """
    Tokenize and format a dataset
    """
    stoi = vocab.get_stoi()
    def tokenize(batch):
        text = unidecode.unidecode(batch['text'].lower())
        tokens = tokenizer(text)

        token_ids = [stoi[t] for t in tokens]
        return {"input_ids": token_ids}

    return dataset.map(tokenize, batched=False)

def format(dataset):
    """
    Format dataset
    """
    dataset = dataset.map(lambda examples: {'labels': examples['label']})
    return dataset

def get_dataloader(dataset, batch_size, pad_idx):
    """
    Get a dataloader for a dataset
    """

    def collate_batch(batch):

        labels = [t["labels"].tolist() if type(t["labels"]) is torch.Tensor else t["labels"] for t in batch]
        input_ids = [t["input_ids"] for t in batch]
        # Return text, text_lens, labels
        text = pad_sequence(input_ids, padding_value=pad_idx, batch_first=True)
        lens = torch.tensor([len(t) for t in input_ids])
        labels = torch.tensor(labels)
        return text, lens, labels

    return DataLoader(dataset, batch_size=batch_size, collate_fn=collate_batch)

def build_dataloaders(train_dataset, dev_dataset, test_dataset, tokenizer, vocab, batch_size, format_dataset=None):
    """
    Build dataloaders
    """

    train_dataset = tokenize_dataset(train_dataset, tokenizer, vocab)
    dev_dataset = tokenize_dataset(dev_dataset, tokenizer, vocab)
    test_dataset = tokenize_dataset(test_dataset, tokenizer, vocab)

    if format_dataset:
        train_dataset = format_dataset(train_dataset)
        dev_dataset = format_dataset(dev_dataset)
        test_dataset = format_dataset(test_dataset)
    else:
        train_dataset = format(train_dataset)
        dev_dataset = format(dev_dataset)
        test_dataset = format(test_dataset)


    for dataset in [train_dataset, dev_dataset, test_dataset]:
        dataset.set_format(type='torch', columns=['input_ids', 'labels'])

    PAD_IDX = vocab.get_stoi()["<pad>"]


    train_dataloader = get_dataloader(train_dataset, batch_size=batch_size, pad_idx=PAD_IDX)
    dev_dataloader = get_dataloader(dev_dataset, batch_size=batch_size, pad_idx=PAD_IDX)
    test_dataloader = get_dataloader(test_dataset, batch_size=batch_size, pad_idx=PAD_IDX)

    return train_dataloader, dev_dataloader, test_dataloader


def train_rnn_model(train_dataset, dev_dataset, test_dataset, lang, id2label, embeddings_path, batch_size=32, epochs=5, bidirectional=False, rnn_units=512, format_dataset=None, **kwargs):
    """
    Train an RNN model
    """

    tokenizer = get_tokenizer("spacy", lang)
    vocab = build_vocab(concatenate_datasets([train_dataset, dev_dataset, test_dataset]), tokenizer)

    PAD_IDX = vocab.get_stoi()["<pad>"]

    train_dataloader, dev_dataloader, test_dataloader = build_dataloaders(train_dataset, dev_dataset, test_dataset, tokenizer, vocab, batch_size, format_dataset)

    emb_matrix = build_embedding_matrix(vocab, fasttext.load_model(embeddings_path))

    model = RNNModel(
        vocab_size=len(vocab), embedding_dim=emb_matrix.shape[-1], pad_idx=PAD_IDX, rnn_units=rnn_units, bidirectional=bidirectional, id2label=id2label,
        embedding_matrix=emb_matrix, freeze_embeddings=True, num_labels=len(id2label),
    )

    trainer = pl.Trainer(
        max_epochs=epochs,
        gpus=1
    )
    trainer.fit(model, train_dataloader, dev_dataloader)
    results = EvalResults(
        predictions=torch.cat(trainer.predict(model, test_dataloader)),
        metrics=trainer.test(model, test_dataloader)[0]
    )
    return trainer, results

def train_ffn_model(train_dataset, dev_dataset, test_dataset, lang, id2label, embeddings_path, batch_size=32, epochs=5, hidden_units=512, format_dataset=None, **kwargs):
    """
    Train an RNN model
    """

    tokenizer = get_tokenizer("spacy", lang)
    vocab = build_vocab(concatenate_datasets([train_dataset, dev_dataset, test_dataset]), tokenizer)

    PAD_IDX = vocab.get_stoi()["<pad>"]

    train_dataloader, dev_dataloader, test_dataloader = build_dataloaders(train_dataset, dev_dataset, test_dataset, tokenizer, vocab, batch_size, format_dataset)

    emb_matrix = build_embedding_matrix(vocab, fasttext.load_model(embeddings_path))

    model = FFNModel(
        vocab_size=len(vocab), embedding_dim=emb_matrix.shape[-1], pad_idx=PAD_IDX,
        hidden_units=hidden_units,
        embedding_matrix=emb_matrix, freeze_embeddings=True, num_labels=len(id2label),
        id2label=id2label,
    )

    trainer = pl.Trainer(
        max_epochs=epochs,
        gpus=1
    )
    trainer.fit(model, train_dataloader, dev_dataloader)
    results = EvalResults(
        predictions=torch.cat(trainer.predict(model, test_dataloader)),
        metrics=trainer.test(model, test_dataloader)[0]
    )
    return trainer, results




"""
Run sentiment experiments
"""
import torch
import pandas as pd
import os
import pathlib
from datasets import Dataset, Value, ClassLabel, Features
from sklearn.model_selection import train_test_split
from .preprocessing import preprocess_tweet
from .training import train_model

extra_args = {
    "vinai/bertweet-base": {
        "preprocessing_args": {"user_token": "@USER", "url_token": "HTTPURL"}
    }
}


project_dir = pathlib.Path(os.path.dirname(__file__)).parent
data_dir = os.path.join(project_dir, "data")
sentiment_dir = os.path.join(data_dir, "irony")

id2label = {
    0: 'not ironic',
    1: 'ironic',
}

label2id = {v:k for k, v in id2label.items()}


def load_datasets(lang, data_path=None, limit=None, random_state=20202021, preprocess=True, preprocess_args={}):
    """
    Load sentiment datasets
    """
    features = Features({
        'text': Value('string'),
        'topic': Value('string'),
        'lang': Value('string'),
        'label': ClassLabel(num_classes=2)
    })
    data_path = data_path or os.path.join(sentiment_dir, "irosva_dataset.csv")
    df = pd.read_csv(data_path)
    df["label"] = df["is_ironic"]


    if preprocess:

        preprocess_fn = lambda x: preprocess_tweet(x, lang=lang, **preprocess_args)
        df["text"] = df["text"].apply(preprocess_fn)
    train_df = df[df["split"] == "train"]
    test_df = df[df["split"] == "test"]

    train_df, dev_df = train_test_split(
        train_df, stratify=train_df["label"], random_state=random_state,
        test_size=0.25,
    )

    train_dataset = Dataset.from_pandas(train_df, features=features)
    dev_dataset = Dataset.from_pandas(dev_df, features=features)
    test_dataset = Dataset.from_pandas(test_df, features=features)


    if limit:
        """
        Smoke test
        """
        print("\n\n", f"Limiting to {limit} instances")
        train_dataset = train_dataset.select(range(min(limit, len(train_dataset))))
        dev_dataset = dev_dataset.select(range(min(limit, len(dev_dataset))))
        test_dataset = test_dataset.select(range(min(limit, len(test_dataset))))


    return train_dataset, dev_dataset, test_dataset


def train(
    base_model, lang="es", epochs=5, batch_size=32,
    limit=None, **kwargs
):
    """
    """

    load_extra_args = extra_args[base_model] if base_model in extra_args else {}

    train_dataset, dev_dataset, test_dataset = load_datasets(lang=lang, **load_extra_args)

    kwargs = {
        **kwargs,
        **{
            "id2label" : id2label,
            "epochs": epochs,
            "batch_size": batch_size,
            "limit": limit,
            "lang": lang,
        }
    }

    return train_model(base_model, train_dataset, dev_dataset, test_dataset, **kwargs)



import torch
from pysentimiento.tass import (
    load_datasets as load_tass_datasets, id2label as id2labeltass, label2id as label2idtass,
)
from pysentimiento.training import load_model, train_model
from pysentimiento.baselines.training import train_rnn_model
from pysentimiento.semeval import (
    load_datasets as load_semeval_datasets,
    id2label as id2labelsemeval, label2id as label2idsemeval
)


lang_conf = {
    "es": {
        "load_datasets": load_tass_datasets,
        "id2label": id2labeltass,
        "label2id": label2idtass,
    },
    "en": {
        "load_datasets": load_semeval_datasets,
        "id2label": id2labelsemeval,
        "label2id": label2idsemeval,
    }
}

extra_args = {
    "vinai/bertweet-base": {
        "preprocessing_args": {"user_token": "@USER", "url_token": "HTTPURL"}
    }
}


def train(
    base_model, lang="es", epochs=5, batch_size=32,
    limit=None, **kwargs
):
    """
    """
    load_datasets = lang_conf[lang]["load_datasets"]
    id2label = lang_conf[lang]["id2label"]

    load_extra_args = extra_args[base_model] if base_model in extra_args else {}

    train_dataset, dev_dataset, test_dataset = load_datasets(**load_extra_args)

    kwargs = {
        **kwargs,
        **{
            "id2label" : id2label,
            "epochs": epochs,
            "batch_size": batch_size,
            "limit": limit,
            "lang": lang,
        }
    }

    return train_model(base_model, train_dataset, dev_dataset, test_dataset, **kwargs)




"""
Run hatEval experiments
"""
import pandas as pd
import os
import pathlib
import torch
import numpy as np
import logging
from datasets import Dataset, Value, ClassLabel, Features
from transformers import Trainer
from sklearn.metrics import precision_recall_fscore_support, accuracy_score, f1_score
from .preprocessing import preprocess_tweet, extra_args
from .training import load_model, train_model



logging.basicConfig()

logger = logging.getLogger('pysentimiento')
logger.setLevel(logging.INFO)


project_dir = pathlib.Path(os.path.dirname(__file__)).parent
data_dir = os.path.join(project_dir, "data", "hate")



def load_datasets(lang,
    train_path=None, dev_path=None, test_path=None, limit=None,
    random_state=2021, preprocess=True, preprocessing_args={} ):
    """
    Load emotion recognition datasets
    """

    train_path = train_path or os.path.join(data_dir, f"hateval2019_{lang}_train.csv")
    dev_path = dev_path or os.path.join(data_dir, f"hateval2019_{lang}_dev.csv")
    test_path = test_path or os.path.join(data_dir, f"hateval2019_{lang}_test.csv")

    logger.info(f"Train path = {train_path}")
    logger.info(f"Dev path = {dev_path}")
    logger.info(f"Test path = {test_path}")


    train_df = pd.read_csv(train_path)
    dev_df = pd.read_csv(dev_path)
    test_df = pd.read_csv(test_path)

    if preprocess:
        preprocess_fn = lambda x: preprocess_tweet(x, lang=lang, **preprocessing_args)

        for df in [train_df, dev_df, test_df]:
            df["text"] = df["text"].apply(preprocess_fn)


    features = Features({
        'id': Value('int64'),
        'text': Value('string'),
        'HS': ClassLabel(num_classes=2, names=["OK", "HATEFUL"]),
        'TR': ClassLabel(num_classes=2, names=["GROUP", "INDIVIDUAL"]),
        "AG": ClassLabel(num_classes=2, names=["NOT AGGRESSIVE", "AGGRESSIVE"])
    })

    train_dataset = Dataset.from_pandas(train_df, features=features)
    dev_dataset = Dataset.from_pandas(dev_df, features=features)
    test_dataset = Dataset.from_pandas(test_df, features=features)


    if limit:
        """
        Smoke test
        """
        print("\n\n", f"Limiting to {limit} instances")
        train_dataset = train_dataset.select(range(min(limit, len(train_dataset))))
        dev_dataset = dev_dataset.select(range(min(limit, len(dev_dataset))))
        test_dataset = test_dataset.select(range(min(limit, len(test_dataset))))



    return train_dataset, dev_dataset, test_dataset

def _get_b_metrics(preds, labels):
    ret = {}

    f1s = []
    precs = []
    recalls = []
    original_preds = preds.copy()

    preds[:, 1] = preds[:, 0] & preds[:, 1]
    preds[:, 2] = preds[:, 0] & preds[:, 2]

    for i, cat in enumerate(["HS", "TR", "AG"]):
        cat_labels, cat_preds = labels[:, i], preds[:, i]


        precision, recall, f1, _ = precision_recall_fscore_support(
            cat_labels, cat_preds, average='binary', zero_division=0,
        )

        f1s.append(f1)
        precs.append(precision)
        recalls.append(recall)

        ret[cat.lower()+"_f1"] = f1
        ret[cat.lower()+"_precision"] = precision
        ret[cat.lower()+"_recall"] = recall



    neg_hs_f1_score = f1_score(1-(preds[:, 0] > 0), 1 - labels[:, 0])

    ret["macro_hs_f1_score"] = (f1s[0] + neg_hs_f1_score) / 2
    #
    # We calculate EMR in a gated way
    # Block TR and AG if HS is False
    #
    ret["emr_no_gating"] = accuracy_score(labels, original_preds)
    ret["emr"] = accuracy_score(labels, preds)

    ret["macro_f1"] = torch.Tensor(f1s).mean()
    ret["macro_precision"] = torch.Tensor(precs).mean()
    ret["macro_recall"] = torch.Tensor(recalls).mean()

    return ret

def get_task_b_metrics(predictions):

    outputs = predictions.predictions
    labels = predictions.label_ids

    return _get_b_metrics(outputs > 0, labels)
# Maps combinations to classes

combinatorial_mapping = {
    (0, 0, 0): 0, # not hateful
    (1, 0, 0): 1, # hs, not tr, not ag
    (1, 0, 1): 2, # hs, not tr, ag
    (1, 1, 0): 3, # hs, tr    , not ag
    (1, 1, 1): 4, # hs, tr    , ag
}

inverse_combinatorial_mapping = {v:k for k, v in combinatorial_mapping.items()}

def get_combinatorial_metrics(predictions):
    outputs = predictions.predictions
    labels = predictions.label_ids

    preds = outputs.argmax(1)

    normalized_preds = np.array([inverse_combinatorial_mapping[k] for k in preds])
    normalized_labels = np.array([inverse_combinatorial_mapping[k] for k in labels])
    return _get_b_metrics(normalized_preds, normalized_labels)




class HierarchicalTrainer(Trainer):
    """
    Hierarchical Cross Entropy loss
    """

    def __init__(self, gamma=0, *args, **kwargs):
        """
        Gamma is the hyperparameter of this loss

        B(y) = (1 - y) γ + y * 1

        such that

        L(y, ypred) = L(y_HS, ypred_HS) + B(y_HS) (L(y_TR, ypred_TR) + L(y_AG, ypred_AG))

        If γ = 1, this is equal to standard sum of binary cross-entropies
        If equals to zero, it only sums the losses of the second-stage variables if the previous is one
        """
        super().__init__(*args, **kwargs)
        self.gamma = gamma

    def compute_loss(self, model, inputs, return_outputs=False):
        labels = inputs.pop("labels")
        outputs = model(**inputs)
        logits = outputs.logits

        loss_fct = torch.nn.BCEWithLogitsLoss(reduction='none')
        unmasked_loss = loss_fct(logits, labels)
        # Expand to two columns, as this is the dimension of the second stage
        mask = labels[:, 0].view(-1, 1).expand(labels.shape[0], 2).clone()
        # Esto está mal porque floats pero bueno
        mask[mask < 1] = self.gamma

        first_stage_loss = unmasked_loss[:, 0]
        # Mask only the second stage
        second_stage_loss = (unmasked_loss[:, 1:] * mask).sum(1)
        loss = (first_stage_loss + second_stage_loss).sum()
        return (loss, outputs) if return_outputs else loss




def train(
    base_model, lang, epochs=5, batch_size=32,
    warmup_ratio=.1, limit=None, accumulation_steps=1, task_b=True, class_weight=None,
    hierarchical=False, gamma=.0, dev=False, metric_for_best_model="macro_f1",
    combinatorial=False, **kwargs,
    ):
    """
    Train function

    Arguments:
    ---------

    task_b: bool (default False)
        If true, trains model for task_b

    combinatorial: bool (default False)
        If task_b true, whether to train a model in a combinatorial fashion

        That is, instead of training a different output for each predicted label,
        train a classifier for 5 possible combinations:
            0: not hateful
            1: HS, not TR, not AG
            2: HS, not TR, AG
            3: HS, TR, not AG
            4: HS, TR, AG
    """

    train_dataset, dev_dataset, test_dataset = load_datasets(
        lang=lang,
        preprocessing_args=extra_args.get(base_model, {})
    )


    if dev:
        test_dataset = dev_dataset

    if limit:
        """
        Smoke test
        """
        print("\n\n", f"Limiting to {limit} instances")
        train_dataset = train_dataset.select(range(limit))
        dev_dataset = dev_dataset.select(range(limit))
        test_dataset = test_dataset.select(range(limit))

    trainer_class = None
    metrics_fun = None
    if task_b:
        if combinatorial:
            id2label = {
                0: "not hateful",
                1: "hateful, not tr, not ag",
                2: "hateful, not tr, ag",
                3: "hateful, targeted, not ag",
                4: "hateful, targeted, ag",
            }

            metrics_fun = get_combinatorial_metrics
        else:
            metrics_fun = get_task_b_metrics
            id2label = {
                0: "hateful",
                1: "targeted",
                2: "aggressive",
            }

        trainer_class = (lambda *args, **kwargs: HierarchicalTrainer(*args, gamma=gamma, **kwargs)) if hierarchical else None
    else:
        metrics_fun = None
        id2label = {
            0: 'ok',
            1: 'hateful',
        }


    label2id = {v:k for k, v in id2label.items()}


    problem_type = "multi_label_classification" if (task_b and not combinatorial) else "single_label_classification"

    labels_order = ["HS", "TR", "AG"]
    def format_dataset(dataset):
        def get_labels(examples):
            if task_b:
                if combinatorial:
                    # Convert to a single label
                    return {'labels': combinatorial_mapping[
                        tuple(examples[k] for k in labels_order)
                    ]}
                return {'labels': torch.Tensor([examples[k] for k in labels_order])}
            else:
                return {'labels': examples["HS"]}
        dataset = dataset.map(get_labels)
        return dataset

    if class_weight:
        class_weight = torch.Tensor([train_dataset[k] for k in labels_order])
        class_weight = 1 / (2* class_weight.mean(1))


    return train_model(
        base_model, train_dataset, dev_dataset, test_dataset, id2label,
        format_dataset=format_dataset, lang=lang,
        epochs=epochs, batch_size=batch_size, class_weight=class_weight,
        warmup_ratio=warmup_ratio, accumulation_steps=accumulation_steps,
        metrics_fun=metrics_fun, trainer_class=trainer_class, metric_for_best_model=metric_for_best_model,
        **kwargs
    )



"""
NER for LinCE dataset
"""
import numpy as np
from .ner import tokenize_and_align_labels
from ..preprocessing import preprocess_tweet
from datasets import load_dataset, load_metric
from seqeval.metrics import f1_score
from ..training import train_model


id2label =[
   "negative",
   "neutral",
   "positive",
]

label2id = {v:k for k,v in enumerate(id2label)}


def load_datasets(lang="es", preprocess=True, preprocessing_args={}):
    """
    Load NER datasets
    """

    lince = load_dataset("lince", "sa_spaeng")
    # This hack is because of seqeval only working with BIO tags
    lince = lince.map(
        lambda x: {
            "text": " ".join(x["words"]),
            "label": label2id.get(x["sa"], None),
        }
    )

    if preprocess:
        preprocess_fn = lambda x: preprocess_tweet(x, lang=lang, **preprocessing_args)

        lince = lince.map(
            lambda x: {"text": preprocess_fn(x["text"])}
        )

    return lince["train"], lince["validation"], lince["test"]


def train(
    base_model, lang, epochs=5,
    metric_for_best_model="macro_f1",
    **kwargs):

    train_dataset, dev_dataset, test_dataset = load_datasets(
        lang=lang
    )

    return train_model(
        base_model,
        train_dataset=train_dataset, dev_dataset=dev_dataset, test_dataset=dev_dataset,
        id2label=id2label, lang=lang, epochs=epochs,
        # Custom stuff for this thing to work
        metric_for_best_model=metric_for_best_model,
        **kwargs
    )


import torch
import torch.nn.functional as F
from torch import nn
import pytorch_lightning as pl
from pysentimiento.metrics import get_metrics


class BaseModel(pl.LightningModule):
    def __init__(self, id2label):
        super().__init__()
        self.problem_type = None
        self.id2label = id2label
        self.loss_fun = None

    def _set_loss_and_problem_type(self, y):
        if not self.problem_type:
            if y.dim() > 1 and y.shape[-1] > 1:
                self.problem_type = "multi_label_classification"
                self.loss_fun = F.binary_cross_entropy_with_logits
            else:
                self.problem_type = "single_label_classification"
                self.loss_fun = F.cross_entropy

    def training_step(self, batch, batch_idx):
        x, lens, y = batch

        self._set_loss_and_problem_type(y)

        outs = self.forward(x, lens)
        loss = self.loss_fun(outs, y)
        self.log('train_loss', loss, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        x, lens, y = batch
        self._set_loss_and_problem_type(y)

        outs = self.forward(x, lens)
        loss = self.loss_fun(outs, y)
        preds = outs.cpu()
        metrics = get_metrics(preds, y.cpu(), self.id2label)
        self.log('val_loss', loss, prog_bar=True, on_epoch=True)

        for k, v in metrics.items():
            self.log("val_"+k, v, prog_bar=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        x, lens, y = batch
        outs = self.forward(x, lens)
        preds = outs.cpu()
        metrics = get_metrics(preds, y.cpu(), self.id2label)

        for k, v in metrics.items():
            self.log("test_"+k, v, prog_bar=True, on_epoch=True)

    def predict_step(self, batch, batch_idx):
        x, lens, y = batch
        return self.forward(x, lens)

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        return optimizer


class RNNModel(BaseModel):
    def __init__(self, vocab_size, embedding_dim, pad_idx, rnn_units, num_labels, id2label, num_layers=1,
                 bidirectional=False, dropout=0.25, embedding_matrix=None, freeze_embeddings=True):

        super().__init__(id2label)

        if embedding_matrix is not None:
            self.embedding = nn.Embedding.from_pretrained(
                embedding_matrix, padding_idx=pad_idx,
                freeze=freeze_embeddings
            )
        else:
            self.embedding = nn.Embedding(
                vocab_size, embedding_dim,
                padding_idx = pad_idx)

        self.rnn = nn.GRU(embedding_dim,
                   rnn_units,
                   num_layers=num_layers,
                   bidirectional=bidirectional, batch_first=True)

        self.dropout = nn.Dropout(dropout)

        factor = 2 if bidirectional else 1

        self.fc = nn.Linear(rnn_units * factor, num_labels)

    def forward(self, text, text_lens):
        #text = [batch_size, text len]
        #permuted = text.permute(1, 0)
        # permuted shape [batch_size, sent len]
        embedded = self.embedding(text)
        packed_embedded = nn.utils.rnn.pack_padded_sequence(
            # WTF no sé por qué hago esto de cpu
            embedded, text_lens.to("cpu"), batch_first=True, enforce_sorted=False)

        packed_output, _ = self.rnn(packed_embedded)
        # hidden is the last state of the

        output, output_lengths = nn.utils.rnn.pad_packed_sequence(packed_output)

        output = self.dropout(output)
        # output is shape [seq, batch, hid]
        s = output.permute(1, 0, 2)
        # now [batch, seq, hid]
        mean = s.sum(dim=1) / text_lens.view(-1, 1)

        return self.fc(mean)

class FFNModel(BaseModel):
    def __init__(self, vocab_size, embedding_dim, pad_idx, hidden_units, num_labels, id2label, dropout=0.25, embedding_matrix=None, freeze_embeddings=True):
        super().__init__(id2label)

        if embedding_matrix is not None:
            self.embedding = nn.Embedding.from_pretrained(
                embedding_matrix, padding_idx=pad_idx,
                freeze=freeze_embeddings
            )
        else:
            self.embedding = nn.Embedding(
                vocab_size, embedding_dim,
                padding_idx = pad_idx)

        self.encoder = nn.Linear(embedding_dim, hidden_units)

        self.dropout = nn.Dropout(dropout)

        self.fc = nn.Linear(hidden_units, num_labels)

    def forward(self, text, text_lens):
        #text = [batch_size, text len]
        #permuted = text.permute(1, 0)
        # permuted shape [batch_size, sent len]
        embedded = self.embedding(text)
        mean_embedding = embedded.mean(dim=1)

        encoded = F.relu(self.encoder(mean_embedding))
        encoded = self.dropout(encoded)
        return self.fc(encoded)



import torch
import os
import torch
import tempfile

from transformers.utils.dummy_pt_objects import AutoModelForTokenClassification
from .metrics import compute_metrics
from transformers import (
    AutoModelForSequenceClassification, AutoTokenizer, DataCollatorWithPadding,
    Trainer, TrainingArguments
)
from .baselines.training import train_rnn_model, train_ffn_model
from .preprocessing import special_tokens

dont_add_tokens = {
    "vinai/bertweet-base"
}




def load_model(
    base_model, id2label, label2id, max_length=128, auto_class=AutoModelForSequenceClassification):
    """
    Loads model and tokenizer
    """
    print(f"Loading model {base_model}")
    model = auto_class.from_pretrained(
        base_model, return_dict=True, num_labels=len(id2label)
    )

    tokenizer = AutoTokenizer.from_pretrained(base_model)
    tokenizer.model_max_length = max_length

    model.config.id2label = id2label
    model.config.label2id = label2id

    if base_model not in dont_add_tokens:
        tokenizer.add_tokens(special_tokens)
        model.resize_token_embeddings(len(tokenizer))
    return model, tokenizer


class MultiLabelTrainer(Trainer):
    """
    Multilabel and class weighted trainer
    """
    def __init__(self, class_weight, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.class_weight = class_weight

    def compute_loss(self, model, inputs, return_outputs=False):
        labels = inputs.pop("labels")
        outputs = model(**inputs)
        logits = outputs.logits
        if self.model.config.problem_type == "multi_label_classification":
            loss_fct = torch.nn.BCEWithLogitsLoss(pos_weight=self.class_weight)
        else:
            loss_fct = torch.nn.CrossEntropyLoss(weight=self.class_weight)
        loss = loss_fct(logits, labels)
        return (loss, outputs) if return_outputs else loss

def train_huggingface(
    model, tokenizer, train_dataset, dev_dataset, test_dataset, id2label,
    epochs=5, batch_size=32, accumulation_steps=1, format_dataset=None, eval_batch_size=32, use_dynamic_padding=True, class_weight=None, group_by_length=True, warmup_ratio=.1, trainer_class=None, load_best_model_at_end=True, metrics_fun=None, metric_for_best_model="macro_f1", data_collator_class=DataCollatorWithPadding, tokenize_fun=None,
    **kwargs):
    """
    Run experiments experiments
    """
    padding = False if use_dynamic_padding else 'max_length'
    def tokenize(batch, tokenizer):
        return tokenizer(batch['text'], padding=padding, truncation=True)

    if tokenize_fun is None:
        tokenize_fun = tokenize

    _tokenize_fun = lambda x: tokenize_fun(x, tokenizer)

    device = "cuda" if torch.cuda.is_available() else "cpu"

    train_dataset = train_dataset.map(_tokenize_fun, batched=True, batch_size=batch_size)
    dev_dataset = dev_dataset.map(_tokenize_fun, batched=True, batch_size=eval_batch_size)
    test_dataset = test_dataset.map(_tokenize_fun, batched=True, batch_size=eval_batch_size)

    if use_dynamic_padding:
        data_collator = data_collator_class(tokenizer, padding="longest")
    else:
        if not format_dataset:
            raise ValueError("Must provide format_dataset if not using dynamic padding")

    if format_dataset:
        train_dataset = format_dataset(train_dataset)
        dev_dataset = format_dataset(dev_dataset)
        test_dataset = format_dataset(test_dataset)

    output_path = tempfile.mkdtemp(prefix="pysentimiento")
    training_args = TrainingArguments(
        output_dir=output_path,
        num_train_epochs=epochs,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=eval_batch_size,
        gradient_accumulation_steps=accumulation_steps,
        warmup_ratio=warmup_ratio,
        evaluation_strategy="epoch",
        save_strategy="epoch",
        do_eval=False,
        weight_decay=0.01,
        logging_dir='./logs',
        load_best_model_at_end=load_best_model_at_end,
        metric_for_best_model=metric_for_best_model,
        group_by_length=group_by_length,
        **kwargs,
    )

    trainer_args = {
        "model": model,
        "args": training_args,
        "compute_metrics": metrics_fun,
        "train_dataset": train_dataset,
        "eval_dataset": dev_dataset,
        "data_collator": data_collator,
        "tokenizer": tokenizer,
    }

    if class_weight is not None:

        class_weight = class_weight.to(device)
        print(f"Using class weight = {class_weight}")
        trainer_class = MultiLabelTrainer
        trainer_args["class_weight"]=class_weight,
    else:
        trainer_class = trainer_class or Trainer

    trainer = trainer_class(**trainer_args)

    trainer.train()

    test_results = trainer.predict(test_dataset)

    os.system(f"rm -Rf {output_path}")

    return trainer, test_results


def train_model(
    base_model, train_dataset, dev_dataset, test_dataset, id2label,
    lang, limit=None, max_length=128, metrics_fun=None,
    auto_class=AutoModelForSequenceClassification,
    **kwargs):
    """
    Base function
    """
    if limit:
        """
        Smoke test
        """
        print("\n\n", f"Limiting to {limit} instances")
        train_dataset = train_dataset.select(range(limit))
        dev_dataset = dev_dataset.select(range(limit))
        test_dataset = test_dataset.select(range(limit))

    if type(id2label) is list:
        id2label = {i: label for i, label in enumerate(id2label)}

    label2id = {v:k for k,v in id2label.items()}

    if not metrics_fun:
        metrics_fun = lambda x: compute_metrics(x, id2label=id2label)

    if base_model == "rnn":
        return train_rnn_model(
            train_dataset, dev_dataset, test_dataset, lang=lang, id2label=id2label, metrics_fun=metrics_fun,
            **kwargs
        )
    elif base_model == "ffn":
        return train_ffn_model(
            train_dataset, dev_dataset, test_dataset, lang=lang, id2label=id2label, metrics_fun=metrics_fun,
            **kwargs
        )
    else:
        """
        Transformer classifier
        """
        device = "cuda" if torch.cuda.is_available() else "cpu"
        model, tokenizer = load_model(
            base_model, label2id=label2id, id2label=id2label,
            max_length=max_length, auto_class=auto_class,
        )

        model = model.to(device)
        model.train()

        return train_huggingface(
            model, tokenizer, train_dataset, dev_dataset, test_dataset, id2label,
            metrics_fun=metrics_fun, **kwargs
        )



import torch
from torchtext.vocab import build_vocab_from_iterator
import unidecode

def build_vocab(dataset, tokenizer):
    """
    Build vocabulary 

    Arguments:

    dataset: datasets.Dataset
        The dataset to extract tokens from

    tokenizer: torchtext.tokenizer
        Tokenizer
    """
    def get_tokens():
        for example in dataset:
            yield tokenizer(unidecode.unidecode(example["text"].lower()))


    vocab = build_vocab_from_iterator(get_tokens(), specials=["<unk>", "<pad>"])
    vocab.set_default_index(vocab["<unk>"])

    return vocab

def build_embedding_matrix(vocab, fasttext_model):
    """
    Builds embedding matrix from a fasttext model

    """
    DIM = fasttext_model.get_word_vector("random").shape[0]
    emb_matrix = torch.randn(len(vocab), DIM)
    stoi = vocab.get_stoi()
    itos = vocab.get_itos()
    UNK_IDX = stoi["<unk>"]
    PAD_IDX = stoi["<pad>"]

    itos = vocab.get_itos()

    # emb_matrix[UNK_IDX] = 0
    emb_matrix[PAD_IDX] = 0

    for i, word in enumerate(itos):
        if i == UNK_IDX or i == PAD_IDX:
            # Let them unmodified
            pass
        else:
            emb_matrix[i] = torch.tensor(fasttext_model.get_word_vector(word))
    return emb_matrix



"""
Auxiliary functions for pysentimiento.
"""
import json


def clean_key(k):
    return k.split("_", 1)[1]

def load_evaluation(path):
    """
    Load evaluation file generated from benchmarking with bin/train.py
    """
    with open(path) as f:
        model_evaluation = json.load(f)
        clean_evaluations = []
        for task in model_evaluation["evaluations"].keys():
            task_evaluations = model_evaluation["evaluations"][task]
            clean_evaluations = [
                {clean_key(k): v for k, v in ev.items()}
                for ev in task_evaluations
            ]

            model_evaluation["evaluations"][task] = clean_evaluations

        return model_evaluation


import torch
from sklearn.metrics import precision_recall_fscore_support, accuracy_score


def compute_metrics(preds, id2label):
    """
    Compute metrics for Trainer
    """
    labels = preds.label_ids

    return get_metrics(preds.predictions, labels, id2label)


def get_metrics(preds, labels, id2label):
    ret = {}

    f1s = []
    precs = []
    recalls = []

    is_multi_label = len(labels.shape) > 1 and labels.shape[-1] > 1

    if not is_multi_label:
        preds = preds.argmax(-1)

    for i, cat in id2label.items():

        if is_multi_label:
            cat_labels, cat_preds = labels[:, i], preds[:, i]

            cat_preds = cat_preds > 0
        else:
            cat_labels, cat_preds = labels == i, preds == i

        precision, recall, f1, _ = precision_recall_fscore_support(
            cat_labels, cat_preds, average='binary', zero_division=0,
        )

        f1s.append(f1)
        precs.append(precision)
        recalls.append(recall)

        ret[cat.lower()+"_f1"] = f1
        ret[cat.lower()+"_precision"] = precision
        ret[cat.lower()+"_recall"] = recall

    if not is_multi_label:
        _, _, micro_f1, _ = precision_recall_fscore_support(
            labels, preds, average="micro"
        )
        ret["micro_f1"] = micro_f1
        ret["acc"] = accuracy_score(labels, preds)
    else:
        ret["emr"] = accuracy_score(labels, preds > 0)

    ret["macro_f1"] = torch.Tensor(f1s).mean()
    ret["macro_precision"] = torch.Tensor(precs).mean()
    ret["macro_recall"] = torch.Tensor(recalls).mean()


    return ret



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysentimiento', 'pysentimiento.baselines', 'pysentimiento.lince']

package_data = \
{'': ['*']}

install_requires = \
['datasets>=1.13.3,<2.0.0',
 'emoji>=1.6.1,<2.0.0',
 'sklearn>=0.0,<0.1',
 'torch>=1.9.0,<2.0.0',
 'transformers==4.13']

setup_kwargs = {
    'name': 'pysentimiento',
    'version': '0.4.2',
    'description': 'A Transformer-based library for SocialNLP tasks',
    'long_description': '# pysentimiento: A Python toolkit for Sentiment Analysis and Social NLP tasks\n\n\n![Tests](https://github.com/finiteautomata/pysentimiento/workflows/run_tests/badge.svg)\n\nA Transformer-based library for SocialNLP tasks.\n\nCurrently supports:\n\n- Sentiment Analysis (Spanish, English)\n- Emotion Analysis (Spanish, English)\n- Hate Speech Detection (Spanish, English)\n- Named Entity Recognition (Spanish + English)\n- POS Tagging (Spanish + English)\n\n\nJust do `pip install pysentimiento` and start using it:\n\n## Getting Started\n\n[![Test it in Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/pysentimiento/pysentimiento/blob/master/notebooks/PySentimiento_Sentiment_Analysis_in_Spanish.ipynb)\n\n```python\nfrom pysentimiento import create_analyzer\nanalyzer = create_analyzer(task="sentiment", lang="es")\n\nanalyzer.predict("Qué gran jugador es Messi")\n# returns AnalyzerOutput(output=POS, probas={POS: 0.998, NEG: 0.002, NEU: 0.000})\nanalyzer.predict("Esto es pésimo")\n# returns AnalyzerOutput(output=NEG, probas={NEG: 0.999, POS: 0.001, NEU: 0.000})\nanalyzer.predict("Qué es esto?")\n# returns AnalyzerOutput(output=NEU, probas={NEU: 0.993, NEG: 0.005, POS: 0.002})\n\nanalyzer.predict("jejeje no te creo mucho")\n# AnalyzerOutput(output=NEG, probas={NEG: 0.587, NEU: 0.408, POS: 0.005})\n"""\nEmotion Analysis in English\n"""\n\nanalyzer = create_analyzer(task="emotion", lang="en")\n\nemotion_analyzer.predict("yayyy")\n# returns AnalyzerOutput(output=joy, probas={joy: 0.723, others: 0.198, surprise: 0.038, disgust: 0.011, sadness: 0.011, fear: 0.010, anger: 0.009})\nemotion_analyzer.predict("fuck off")\n# returns AnalyzerOutput(output=anger, probas={anger: 0.798, surprise: 0.055, fear: 0.040, disgust: 0.036, joy: 0.028, others: 0.023, sadness: 0.019})\n\n"""\nHate Speech (misogyny & racism)\n"""\nhate_speech_analyzer = create_analyzer(task="hate_speech", lang="es")\n\nhate_speech_analyzer.predict("Esto es una mierda pero no es odio")\n# returns AnalyzerOutput(output=[], probas={hateful: 0.022, targeted: 0.009, aggressive: 0.018})\nhate_speech_analyzer.predict("Esto es odio porque los inmigrantes deben ser aniquilados")\n# returns AnalyzerOutput(output=[\'hateful\'], probas={hateful: 0.835, targeted: 0.008, aggressive: 0.476})\n\nhate_speech_analyzer.predict("Vaya guarra barata y de poca monta es XXXX!")\n# returns AnalyzerOutput(output=[\'hateful\', \'targeted\', \'aggressive\'], probas={hateful: 0.987, targeted: 0.978, aggressive: 0.969})\n```\n\nAlso, you might use pretrained models directly with [`transformers`](https://github.com/huggingface/transformers) library.\n\n```python\nfrom transformers import AutoTokenizer, AutoModelForSequenceClassification\n\ntokenizer = AutoTokenizer.from_pretrained("pysentimiento/robertuito-sentiment-analysis")\n\nmodel = AutoModelForSequenceClassification.from_pretrained("pysentimiento/robertuito-sentiment-analysis")\n```\n\n## Preprocessing\n\n`pysentimiento` features a tweet preprocessor specially suited for tweet classification with transformer-based models.\n\n```python\nfrom pysentimiento.preprocessing import preprocess_tweet\n\n# Replaces user handles and URLs by special tokens\npreprocess_tweet("@perezjotaeme debería cambiar esto http://bit.ly/sarasa") # "@usuario debería cambiar esto url"\n\n# Shortens repeated characters\npreprocess_tweet("no entiendo naaaaaaaadaaaaaaaa", shorten=2) # "no entiendo naadaa"\n\n# Normalizes laughters\npreprocess_tweet("jajajajaajjajaajajaja no lo puedo creer ajajaj") # "jaja no lo puedo creer jaja"\n\n# Handles hashtags\npreprocess_tweet("esto es #UnaGenialidad")\n# "esto es una genialidad"\n\n# Handles emojis\npreprocess_tweet("🎉🎉", lang="en")\n# \'emoji party popper emoji emoji party popper emoji\'\n```\n\n## Trained models so far\n\nCheck [CLASSIFIERS.md](CLASSIFIERS.md) for details on the reported performances of each model.\n\n\n## Instructions for developers\n\n0. Clone and install\n\n```\ngit clone https://github.com/pysentimiento/pysentimiento\npip install poetry\npoetry shell\npoetry install\n```\n\n1. Get the data and put it under `data/`\n\nOpen an issue or email us if you are not able to get the it.\n\n2. Run script to train models\n\nCheck [TRAIN.md](TRAIN.md) for further information on how to train your models\n\n3. Upload models to Huggingface\'s Model Hub\n\nCheck ["Model sharing and upload"](https://huggingface.co/transformers/model_sharing.html) instructions in `huggingface` docs.\n\n## License\n\n`pysentimiento` is an open-source library. However, please be aware that models are trained with third-party datasets and are subject to their respective licenses, many of which are for non-commercial use\n\n1. [TASS Dataset license](http://tass.sepln.org/tass_data/download.php) (License for Sentiment Analysis in Spanish, Emotion Analysis in Spanish & English)\n2. [SEMEval 2017 Dataset license](https://www.dropbox.com/s/byzr8yoda6bua1b/2017_English_final.zip?file_subpath=%2F2017_English_final%2FDOWNLOAD%2FREADME.txt) (Sentiment Analysis in English)\n\n3. [LinCE Datasets](https://ritual.uh.edu/lince/datasets) (License for NER & POS tagging)\n\n## Suggestions and bugfixes\n\nPlease use the repository [issue tracker](https://github.com/pysentimiento/pysentimiento/issues) to point out bugs and make suggestions (new models, use another datasets, some other languages, etc)\n\n\n## Citation\n\nIf you use `pysentimiento` in your work, please cite [this paper](https://arxiv.org/abs/2106.09462)\n\n```\n@misc{perez2021pysentimiento,\n      title={pysentimiento: A Python Toolkit for Sentiment Analysis and SocialNLP tasks},\n      author={Juan Manuel Pérez and Juan Carlos Giudici and Franco Luque},\n      year={2021},\n      eprint={2106.09462},\n      archivePrefix={arXiv},\n      primaryClass={cs.CL}\n}\n```\n\nAlso, pleace cite related pre-trained models and datasets for the specific models you use:\n\n```bibtex\n\n%%%%%%%%%%%%%%%%%%%%%%%%%%\n% Pretrained models      %\n%%%%%%%%%%%%%%%%%%%%%%%%%%\n% RoBERTuito\n@article{perez2021robertuito,\n  title={RoBERTuito: a pre-trained language model for social media text in Spanish},\n  author={P{\\\'e}rez, Juan Manuel and Furman, Dami{\\\'a}n A and Alemany, Laura Alonso and Luque, Franco},\n  journal={arXiv preprint arXiv:2111.09453},\n  year={2021}\n}\n% BETO\n@article{canete2020spanish,\n  title={Spanish pre-trained bert model and evaluation data},\n  author={Canete, Jos{\\\'e} and Chaperon, Gabriel and Fuentes, Rodrigo and Ho, Jou-Hui and Kang, Hojin and P{\\\'e}rez, Jorge},\n  journal={Pml4dc at iclr},\n  volume={2020},\n  pages={2020},\n  year={2020}\n}\n% BERTweet\n@inproceedings{nguyen2020bertweet,\n  title={BERTweet: A pre-trained language model for English Tweets},\n  author={Nguyen, Dat Quoc and Vu, Thanh and Nguyen, Anh Tuan},\n  booktitle={Proceedings of the 2020 Conference on Empirical Methods in Natural Language Processing: System Demonstrations},\n  pages={9--14},\n  year={2020}\n}\n%%%%%%%%%%%%%%%%%%%%%%%%%%\n% Datasets               %\n%%%%%%%%%%%%%%%%%%%%%%%%%%\n% TASS 2020 (sentiment in Spanish)\n\n@article{garcia2020overview,\n  title={Overview of TASS 2020: introducing emotion detection},\n  author={Garc{\\\'\\i}a-Vegaa, Manuel and D{\\\'\\i}az-Galianoa, Manuel Carlos and Garc{\\\'\\i}a-Cumbrerasa, Miguel {\\\'A} and del Arcoa, Flor Miriam Plaza and Montejo-R{\\\'a}eza, Arturo and Jim{\\\'e}nez-Zafraa, Salud Mar{\\\'\\i}a and C{\\\'a}marab, Eugenio Mart{\\\'\\i}nez and Aguilarc, C{\\\'e}sar Antonio and Antonio, Marco and Cabezudod, Sobrevilla and others},\n  year={2020}\n}\n\n% EmoEvent (Emotion Analysis Spanish & English)\n\n@inproceedings{del2020emoevent,\n  title={EmoEvent: A multilingual emotion corpus based on different events},\n  author={del Arco, Flor Miriam Plaza and Strapparava, Carlo and Lopez, L Alfonso Urena and Mart{\\\'\\i}n-Valdivia, M Teresa},\n  booktitle={Proceedings of the 12th Language Resources and Evaluation Conference},\n  pages={1492--1498},\n  year={2020}\n}\n\n% Hate Speech Detection (Spanish & English)\n\n\n@inproceedings{hateval2019semeval,\n  title={SemEval-2019 Task 5: Multilingual Detection of Hate Speech Against Immigrants and Women in Twitter},\n  author={Basile, Valerio and Bosco, Cristina and Fersini, Elisabetta and Nozza, Debora and Patti, Viviana and Rangel, Francisco and Rosso, Paolo and Sanguinetti, Manuela},\n  booktitle={Proceedings of the 13th International Workshop on Semantic Evaluation (SemEval-2019)},\n  year={2019},\n  publisher= {Association for Computational Linguistics}\n}\n% Sentiment Analysis in English\n\n@article{nakov2019semeval,\n  title={SemEval-2016 task 4: Sentiment analysis in Twitter},\n  author={Nakov, Preslav and Ritter, Alan and Rosenthal, Sara and Sebastiani, Fabrizio and Stoyanov, Veselin},\n  journal={arXiv preprint arXiv:1912.01973},\n  year={2019}\n}\n\n% LinCE (NER & POS Tagging)\n\n@inproceedings{aguilar2020lince,\n  title={LinCE: A Centralized Benchmark for Linguistic Code-switching Evaluation},\n  author={Aguilar, Gustavo and Kar, Sudipta and Solorio, Thamar},\n  booktitle={Proceedings of the 12th Language Resources and Evaluation Conference},\n  pages={1803--1813},\n  year={2020}\n}\n```\n',
    'author': 'Juan Manuel Pérez',
    'author_email': 'jmperez@dc.uba.ar',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pysentimiento/pysentimiento/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
